#include <linux/limits.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/wait.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <string.h>

#ifndef PORT
#define PORT 10000
#endif

#define MAXLINE 1024

void exec_cmd(const char *cmd, const char *file)
{
	char buf[2 * MAXLINE];

	snprintf(buf, 2 * MAXLINE, "%s %s\n", cmd, file);
	if (system(buf)) {
		fprintf(stderr, "system error\n");
	}
}

void child()
{
	char buf[MAXLINE], cmd, *arg;
	size_t n, arg_len, i;
	char *last_file = NULL, *tmp;
	char *pwd = NULL;
	int pwd_len = 16;

	for (;;) {
		if (fgets(buf, MAXLINE, stdin) == NULL)
			break;
		n = strlen(buf);
		if (n < 2)
			continue;
		if (buf[n - 1] != '\n') {
			fprintf(stderr, "line too long, ignored\n");
			continue;
		}

		cmd = buf[0];
		arg = &buf[1];
		arg_len = n - 2; // exclude terminating newline and cmd
		switch (cmd) {
		case 'p': // set password
			if (arg_len != pwd_len) {
				fprintf(stderr, "invalid new password\n");
				break;
			}
			if (pwd == NULL) {
				pwd = malloc(pwd_len);
				if (pwd == NULL) {
					fprintf(stderr, "out of memory\n");
					break;
				}
			}
			memcpy(pwd, arg, pwd_len);
			break;
		case 's': // set password size
			pwd_len = atoi(arg);
			break;
		case 'c': // cat file
			for (i = 0; i < arg_len; i++) {
				if (!isalnum(arg[i]))
					break;
			}
			if (i != arg_len) {
				fprintf(stderr, "invalid arg\n");
				break;
			}
			tmp = malloc(arg_len);
			if (tmp == NULL) {
				fprintf(stderr, "out of memory\n");
				break;
			}
			memcpy(tmp, arg, arg_len);
			free(last_file);
			last_file = tmp;
			exec_cmd("cat", last_file);
			break;
		case 'l':
			exec_cmd("ls", ".");
			break;
		case 'r': // repeat last cat
			exec_cmd("cat", last_file == NULL ? "banner" : last_file);
			break;
		case 'q': // quit
			return;
		default:
			fprintf(stderr, "unknown command\n");
			break;
		}
	}
}

void sigchld(int signo)
{
	int status;
	pid_t pid = waitpid(-1, &status, WNOHANG);
	if (pid > 0) {
		printf("child %ld: ", (long)pid);
		if (WIFEXITED(status)) {
			printf("exited with status: %d\n", WEXITSTATUS(status));
		} else {
			printf("%s\n", strsignal(WTERMSIG(status)));
		}
	}
}

int main()
{
	int lstn;
	int enable;
	struct sockaddr_in lstn_addr;

	lstn = socket(AF_INET, SOCK_STREAM, 0);
	if (lstn < 0) {
		perror("socket");
		return 1;
	}
	enable = 1;
	if (setsockopt(lstn, SOL_SOCKET, SO_REUSEADDR, &enable, sizeof(enable)) < 0) {
		perror("setsockopt");
		return 1;
	}
	bzero(&lstn_addr, sizeof(lstn_addr));

	lstn_addr.sin_family = AF_INET;
	lstn_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	lstn_addr.sin_port = htons(PORT);

	if (bind(lstn, (struct sockaddr *)&lstn_addr, sizeof(lstn_addr)) < 0) {
		perror("bind");
		return 1;
	}

	if (listen(lstn, 10) < 0) {
		perror("listen");
		return 1;
	}
	printf("Listening on port %d\n", PORT);

	signal(SIGCHLD, sigchld);

	for (;;) {
		int con = accept(lstn, NULL, NULL);
		if (con < 0) {
			perror("accept");
			return 1;
		}

		switch (fork()) {
		case -1:
			perror("fork");
			return 1;
		case 0:
			printf("New connection, child %d\n", getpid());

			close(0);
			if (dup(con) < 0)
				goto error;
			close(1);
			if (dup(con) < 0)
				goto error;
			close(2);
			if (dup(con) < 0)
				goto error;
			child();
			exit(0);
			break;
		default:
			close(con);
			break;
		}
	}
	return 0;

error:
	fprintf(stderr, "dup() failed\n");
	exit(1);
}
